import torch
import numpy as np
import torch.nn as nn
import time
import torch.utils.data as Data
from scNovel.scNovel_classifier import classifier
from scNovel.weightsampling import Weighted_Sampling
from scNovel.score_functions import get_ood_scores





def scRare(test = None, reference = None, label = None, processing_unit = 'cuda',score_function="sim",iteration_number=3000):

    '''

    :param test: Input your test dataset to annotate
    :param reference: Input your reference dataset
    :param label: Input label set
    :param processing_unit: Choose which processing unit you would like to use,
    if you choose 'gpu', than the software will try to access CUDA,
    if you choose 'cpu', the software will use CPU as default.
    :return:
    '''

    label.columns = ['Label']
    label.Label.value_counts()
    status_dict = label['Label'].unique().tolist()
    int_label = label['Label'].apply(lambda x: status_dict.index(x))
    label.insert(1,'transformed',int_label)

    X_train = reference.values
    X_test = test.values
    y_train = label['transformed'].values

    dtype = torch.float
    X_train = torch.from_numpy(X_train)
    X_test = torch.from_numpy(X_test)
    y_train = torch.from_numpy(y_train)

    sampler = Weighted_Sampling(y_train)

    #construct pytorch object
    train_data = Data.TensorDataset(X_train, y_train)
    train_loader = Data.DataLoader(dataset=train_data,
                                   batch_size=32,
                                   sampler = sampler,
                                   num_workers=1)


    test_data = Data.TensorDataset(X_test, torch.zeros(len(X_test)))
    test_loader = Data.DataLoader(dataset=test_data,
                                   batch_size=32,
                                   num_workers=1)


    input_size = X_train.shape[1]
    num_class = len(status_dict)
    iteration=iteration_number
    count_iteration=0

    model = classifier(input_size, num_class)

    criterion = nn.CrossEntropyLoss()
    optimizer = torch.optim.AdamW(model.parameters(), lr=1e-4)
    # total_step = len(train_loader)

    print("--------Start annotating----------")
    start = time.perf_counter()

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    if processing_unit == 'cpu':
        device = torch.device('cpu')
    elif processing_unit == 'gpu':
        if torch.cuda.is_available():
            device = torch.device('cuda')
        else:
            device = torch.device('cpu')
            print("No GPUs are available on your server.")

    print("Computational unit be used is:", device)

    model.to(device)

    model.train()
    while count_iteration<iteration:
        for step, (batch_x, batch_y) in enumerate(train_loader):
            batch_x, batch_y = batch_x.to(device), batch_y.to(device)
            outputs = model(batch_x)
            train_loss = criterion(outputs, batch_y)

            # Backward and optimize
            optimizer.zero_grad()
            train_loss.backward()
            optimizer.step()
            count_iteration+=1
            # print(count_iteration)
            if count_iteration>=iteration:
                break

        print("Finish {}/{}".format(float(count_iteration), iteration))
    # print("Finish {}/{}".format(float(count_iteration), iteration))
    print("\n--------Annotation Finished----------")

    model.eval()

    train_score=get_ood_scores(train_loader,model,score_function,device=device)
    test_score = get_ood_scores(test_loader, model, score_function, device=device)

    max_socre=np.max(train_score)

    test_score=(test_score-1.0/num_class)/(max_socre-1.0/num_class)



    return test_score



import numpy as np
import pandas as pd
import scanpy as sc
from warnings import simplefilter

def Scanpy_Obj_IO(test_obj=None, ref_obj=None, label_obj=None, scale = False):
    '''

    :param test_obj: Input Anndata object for cell type annotation
    :param ref_obj: Input reference cell atlas, please note that an csv file required
    :param label_obj: (IF APPLICABLE) If the reference dataset is splited as expression matrix.csv and label.csv, input the label path
    :return: test_matrix (pd.dataframe) for annotation,
    ref_matrix (pd.dataframe) contain reference expression matrix,
    label (pd.dataframe) contain the label list.
    :scale: Whether to scale the reference dataset
    '''

    ref_adata = ref_obj
    # normallize and log reference dataset
    if label_obj is None:
        label = ref_adata[:, ref_adata.n_vars - 1].to_df()
        ref_adata = ref_adata[:, 0:ref_adata.n_vars - 1]

    else:
        label = label_obj

    sc.pp.normalize_total(ref_adata, target_sum=1e4)
    sc.pp.log1p(ref_adata)
    simplefilter(action='ignore', category=FutureWarning)
    gene = ref_adata.var_names & test_obj.var_names

    if 'log1p' in test_obj.uns:
        # test whether the dataset is in log format
        ref = ref_adata[:,gene]
        test_matrix = test_obj.to_df()[gene]
    else:
        sc.pp.log1p(test_obj)
        ref = ref_adata[:,gene]
        test_matrix = test_obj.to_df()[gene]

    if scale:
        sc.pp.scale(ref, max_value=10)
        ref_matrix = ref.to_df()
    else:
        ref_matrix = ref.to_df()


    return test_matrix, ref_matrix, label



def CSV_IO(csv_path:str = None, label_path:str = None, with_label:bool = True):

    '''

    :param csv_path: Input csv path to load expression matrix
    :param label_path: Input csv path to load label
    :param label: whether we have label (whether this is the reference)
    :return:
    '''
    #normalize the reference dataset
    adata = sc.read_csv(csv_path)
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)

    if with_label:
        #load label
        label = pd.read_csv(label_path)

        #turn to label matrix
        label.columns = ['Label']
        status_dict = label['Label'].unique().tolist()
        label['transfromed'] = label['Label'].apply(lambda x: status_dict.index(x))
        label_matrix = label['transfromed'].values

        count_matrix = adata.to_df()
        return count_matrix, label_matrix, status_dict

    else:
        count_matrix = adata.to_df()

        return count_matrix





import torch
import torch.utils.data as Data

def Weighted_Sampling(train_label = None):

    '''

    :param train_label: Input train label (in tensor) to calculate sampling weight
    :return: Pytorch Weighted Sampler
    '''

    class_sample_count = torch.tensor(
        [(train_label == t).sum() for t in torch.unique(train_label, sorted=True)])

    weight = 1. / class_sample_count.float()
    samples_weight = torch.tensor([weight[t] for t in train_label])

    sampler = Data.WeightedRandomSampler(samples_weight, len(samples_weight))

    return sampler




import torch
import torch.nn as nn

class classifier(nn.Module):
    '''
    Deep Neural Network with 3 hidden layers,
    each layer has batch normalization function locate before the activation function (ELU),
    the followed by the dropout function to reduce overfiting the reference and overcome batch effect
    '''
    def __init__(self, input_size, class_num):
        '''
        :param input_size: Input layer unit number (feature number)
        :param class_num: Number of different cell types.
        '''
        if input_size == None or class_num == None:
            raise ValueError("Must declare number of features and number of classes")
        super(classifier, self).__init__()



        if class_num<16:
            self.layer1 = nn.Linear(input_size, 128)
            self.bn1 = nn.BatchNorm1d(128)
            self.elu1 = nn.ELU()
            self.dropout1 = nn.Dropout(0.5)
            self.layer2 = nn.Linear(128, 64)
            self.bn2 = nn.BatchNorm1d(64)
            self.elu2 = nn.ELU()
            self.dropout2 = nn.Dropout(0.5)
            self.layer3 = nn.Linear(64, 32)
            self.bn3 = nn.BatchNorm1d(32)
            self.elu3 = nn.ELU()
            self.dropout3 = nn.Dropout(0.1)
            self.layer4 = nn.Linear(32, class_num)
        elif class_num<64:
            self.layer1 = nn.Linear(input_size, 256)
            self.bn1 = nn.BatchNorm1d(256)
            self.elu1 = nn.ELU()
            self.dropout1 = nn.Dropout(0.5)
            self.layer2 = nn.Linear(256, 128)
            self.bn2 = nn.BatchNorm1d(128)
            self.elu2 = nn.ELU()
            self.dropout2 = nn.Dropout(0.5)
            self.layer3 = nn.Linear(128, 64)
            self.bn3 = nn.BatchNorm1d(64)
            self.elu3 = nn.ELU()
            self.dropout3 = nn.Dropout(0.1)
            self.layer4 = nn.Linear(64, class_num)
        else :
            self.layer1 = nn.Linear(input_size, 512)
            self.bn1 = nn.BatchNorm1d(512)
            self.elu1 = nn.ELU()
            self.dropout1 = nn.Dropout(0.5)
            self.layer2 = nn.Linear(512, 256)
            self.bn2 = nn.BatchNorm1d(256)
            self.elu2 = nn.ELU()
            self.dropout2 = nn.Dropout(0.5)
            self.layer3 = nn.Linear(256, 128)
            self.bn3 = nn.BatchNorm1d(128)
            self.elu3 = nn.ELU()
            self.dropout3 = nn.Dropout(0.1)
            self.layer4 = nn.Linear(128, class_num)
    def forward(self, x):
        '''
        :param x: forward calculation
        :return:
        '''
        out = self.layer1(x)
        out = self.bn1(out)
        out = self.elu1(out)
        out = self.dropout1(out)
        out = self.layer2(out)
        out = self.bn2(out)
        out = self.elu2(out)
        out = self.dropout2(out)
        out = self.layer3(out)
        out = self.bn3(out)
        out = self.elu3(out)
        out = self.dropout3(out)
        # print(torch.max(out))


        out = self.layer4(out)
        return out




import setuptools

setuptools.setup(
    name = "scNovel",
    version = "1.1.0",
    description = "scNovel, a neural network framework, "
                  "provides a fast and reliable tools for scRNA-seq novel rare cell detection.",
    license = "MIT Licence",
    python_requires=">=3.5.0",
    packages=setuptools.find_packages(),
    url = "https://github.com/chuanyang-Zheng/scNovel",
    author = "Chuanyang Zheng",
    author_email = "cyzheng21@cse.cuhk.edu.hk"
)


import numpy as np
import torch
from torch.autograd import Variable
import torch.nn as nn
import torch.nn.functional as F

def get_socre(outputs):
    nnOutputs = outputs.data.cpu()
    nnOutputs = nnOutputs.numpy()
    nnOutputs = nnOutputs - np.max(nnOutputs, axis=1, keepdims=True)
    nnOutputs = np.exp(nnOutputs) / np.sum(np.exp(nnOutputs), axis=1, keepdims=True)

    return nnOutputs

def ODIN(inputs, outputs, model, temper, noiseMagnitude1, device):
    # Calculating the perturbation we need to add, that is,
    # the sign of gradient of cross entropy loss w.r.t. input
    criterion = nn.CrossEntropyLoss()

    maxIndexTemp = np.argmax(outputs.data.cpu().numpy(), axis=1)

    # Using temperature scaling
    outputs = outputs / temper

    labels = Variable(torch.LongTensor(maxIndexTemp).to(device))
    loss = criterion(outputs, labels)
    loss.backward()

    # Normalizing the gradient to binary in {0, 1}
    gradient = torch.ge(inputs.grad.data, 0)
    gradient = (gradient.float() - 0.5) * 2

    # gradient[:,0] = (gradient[:,0] )/(63.0/255.0)
    # gradient[:,1] = (gradient[:,1] )/(62.1/255.0)
    # gradient[:,2] = (gradient[:,2] )/(66.7/255.0)
    # gradient.index_copy_(1, torch.LongTensor([0]).to(device), gradient.index_select(1, torch.LongTensor([0]).to(device)) / (63.0/255.0))
    # gradient.index_copy_(1, torch.LongTensor([1]).to(device), gradient.index_select(1, torch.LongTensor([1]).to(device)) / (62.1/255.0))
    # gradient.index_copy_(1, torch.LongTensor([2]).to(device), gradient.index_select(1, torch.LongTensor([2]).to(device)) / (66.7/255.0))

    # Adding small perturbations to images
    tempInputs = torch.add(inputs.data, -noiseMagnitude1, gradient)
    outputs = model(Variable(tempInputs))
    outputs = outputs / temper
    # Calculating the confidence after adding perturbations
    nnOutputs = outputs.data.cpu()
    nnOutputs = nnOutputs.numpy()
    nnOutputs = nnOutputs - np.max(nnOutputs, axis=1, keepdims=True)
    nnOutputs = np.exp(nnOutputs) / np.sum(np.exp(nnOutputs), axis=1, keepdims=True)

    return nnOutputs






def SEQ(inputs, outputs, model, temper=1000, noiseMagnitude1=0.0014, device=torch.device('cpu')):
    # Calculating the perturbation we need to add, that is,
    # the sign of gradient of cross entropy loss w.r.t. input
    criterion = nn.CrossEntropyLoss()

    maxIndexTemp = np.argmax(outputs.data.cpu().numpy(), axis=1)
    score_ori=outputs
    score_ori=get_socre(score_ori/temper)
    # print("Raw {}".format(np.max(score_ori, 1)))

    # Using temperature scaling
    outputs = outputs / temper

    labels = Variable(torch.LongTensor(maxIndexTemp).to(device))
    loss = criterion(outputs, labels)
    loss.backward()

    # Normalizing the gradient to binary in {0, 1}
    gradient = torch.ge(inputs.grad.data, 0)
    gradient = (gradient.float() - 0.5) * 2



    # Adding small perturbations to images
    tempInputs = torch.add(inputs.data, -noiseMagnitude1, gradient)
    tempInputs =nn.Parameter(tempInputs)
    outputs = model(tempInputs)
    outputs = outputs / temper
    # Calculating the confidence after adding perturbations
    nnOutputs = outputs.data.cpu()
    nnOutputs = nnOutputs.numpy()
    nnOutputs = nnOutputs - np.max(nnOutputs, axis=1, keepdims=True)
    nnOutputs_ori = np.exp(nnOutputs) / np.sum(np.exp(nnOutputs), axis=1, keepdims=True)
    # print("Ori {}".format(np.max(nnOutputs_ori, 1)))



    # Adding small perturbations to images

    # print(torch.max(tempInputs.grad.data))
    loss = criterion(outputs, labels)
    loss.backward()

    # Normalizing the gradient to binary in {0, 1}
    gradient = torch.ge(tempInputs.grad.data, 0)
    gradient = (gradient.float() - 0.5) * 2
    # print(torch.max(tempInputs.grad.data))

    tempInputs = torch.add(tempInputs.data, noiseMagnitude1, gradient)
    outputs = model(Variable(tempInputs))
    outputs = outputs / temper
    # Calculating the confidence after adding perturbations
    nnOutputs = outputs.data.cpu()
    nnOutputs = nnOutputs.numpy()
    nnOutputs = nnOutputs - np.max(nnOutputs, axis=1, keepdims=True)
    nnOutputs_novel = np.exp(nnOutputs) / np.sum(np.exp(nnOutputs), axis=1, keepdims=True)
    # print("Novel {}".format(np.max(nnOutputs_novel, 1)))

    nnOutputs=nnOutputs_ori+nnOutputs_ori-nnOutputs_novel

    return nnOutputs

def SIM(inputs, outputs, model, temper=1000, noiseMagnitude1=0.0014, device=torch.device('cpu')):
    # Calculating the perturbation we need to add, that is,
    # the sign of gradient of cross entropy loss w.r.t. input
    criterion = nn.CrossEntropyLoss()

    maxIndexTemp = np.argmax(outputs.data.cpu().numpy(), axis=1)
    score_ori=outputs
    score_ori=get_socre(score_ori/temper)
    # print("Raw {}".format(np.max(score_ori, 1)))

    # Using temperature scaling
    outputs = outputs / temper

    labels = Variable(torch.LongTensor(maxIndexTemp).to(device))
    loss = criterion(outputs, labels)
    loss.backward()

    # Normalizing the gradient to binary in {0, 1}
    gradient = torch.ge(inputs.grad.data, 0)
    gradient = (gradient.float() - 0.5) * 2


    # Adding small perturbations to images
    tempInputs = torch.add(inputs.data, -noiseMagnitude1, gradient)
    outputs = model(Variable(tempInputs))
    outputs = outputs / temper
    # Calculating the confidence after adding perturbations
    nnOutputs = outputs.data.cpu()
    nnOutputs = nnOutputs.numpy()
    nnOutputs = nnOutputs - np.max(nnOutputs, axis=1, keepdims=True)
    nnOutputs_ori = np.exp(nnOutputs) / np.sum(np.exp(nnOutputs), axis=1, keepdims=True)
    # print("Ori {}".format(np.max(nnOutputs_ori, 1)))



    # Adding small perturbations to images
    tempInputs = torch.add(inputs.data, noiseMagnitude1, gradient)
    outputs = model(Variable(tempInputs))
    outputs = outputs / temper
    # Calculating the confidence after adding perturbations
    nnOutputs = outputs.data.cpu()
    nnOutputs = nnOutputs.numpy()
    nnOutputs = nnOutputs - np.max(nnOutputs, axis=1, keepdims=True)
    nnOutputs_novel = np.exp(nnOutputs) / np.sum(np.exp(nnOutputs), axis=1, keepdims=True)
    # print("Novel {}".format(np.max(nnOutputs_novel, 1)))

    nnOutputs=score_ori+(nnOutputs_ori-score_ori)+score_ori-nnOutputs_novel

    return nnOutputs


# score_via_name={"msp":,"odin":ODIN,"seq":SEQ,"sim":SIM}
def get_ood_scores(loader, net,score_function="sim",T=1000,noise=0.0014,device=torch.device('cpu')):
    _score = []
    _right_score = []
    _wrong_score = []
    _label = []
    _patient=[]
    to_np = lambda x: x.data.cpu().numpy()
    concat = lambda x: np.concatenate(x, axis=0)

    net.eval()
    for step, (batch_x, batch_y) in enumerate(loader):
        data, target = batch_x, batch_y


        # if batch_idx >= ood_num_examples // bs and in_dist is False:
        #     break
        data = data.to(device)
        data = Variable(data, requires_grad=True)

        output = net(data)

        if score_function=="sim":
            odin_score = SIM(data, output, net, temper=T, noiseMagnitude1=noise, device=device)
        elif score_function=="seq":
            odin_score = SIM(data, output, net, temper=T, noiseMagnitude1=noise, device=device)
        elif score_function=="odin":
            odin_score = ODIN(data, output, net, temper=T, noiseMagnitude1=noise, device=device)
        elif score_function=="msp":
            odin_score=to_np(F.softmax(output, dim=1))
        else:
            raise ValueError ("No Such score function {}".format(score_function))
        _score.append(np.max(odin_score, 1))


    return concat(_score).copy()



